/*
 *This files contains the Agent declarationa and Timers
 */

#ifndef __node_num__
#define __node_num__
const int NODENUM=400;
#endif
 
#ifndef __density_h__
#define __density_h__

#include "density_packet.h"
#include "neighbor.h"
#include "config.h"
#include "object.h"
#include "agent.h"
#include "timer-handler.h"
#include "address.h"
#include <lib/bsd-list.h>
#include "densityrouting.h"

class DensityAgent;
 
#define DISC              1
#define ADV               2
#define DST_FORWARDING    3
#define DST_REP           4
#define DELEGATION        5
#define SND_HELLO         6
#define CHECK_NEIGHBORS   7


/*=========================================================================================================
 *                                             TIMERS
 *=========================================================================================================*/
 

/*============================================================
 *  BOUNDARIES IDENTIFICATION PROTOCOL
 *============================================================*/

/*
 *The discovery timer sets the time passing between the broadcast of node identities
 *and the moment when nodes broadcast the number of its neighbors
 */

 class DiscoveryTimer : public TimerHandler {
 public:
    DiscoveryTimer (DensityAgent* a) :TimerHandler(), a_(a) {};
 protected:
    virtual void expire(Event *);
    DensityAgent* a_;
 };


/*
 *When the participation timer fire, the node decides whether to participate
 *to the locality or not
 */
class PartAdvTimer : public TimerHandler {
 public:
    PartAdvTimer (DensityAgent* a) :TimerHandler(), a_(a) {};
 protected:
    virtual void expire(Event *);
    DensityAgent* a_;
 };

/*============================================================
 *        MANAGER ELECTION PROTOCOL                           *
 *============================================================*/

/*======FARTHEST NODES IDENTIFICATION============================*/
/*
 *When the distance reply timer fire, the node replies to the current IN
 */
class DistanceRepTimer : public TimerHandler {
 public:
    DistanceRepTimer (DensityAgent* a) :TimerHandler(), a_(a) {};
 protected:
    virtual void expire(Event *);
    DensityAgent* a_;
 };

/*
 *When the distance forwarding timer fires, the node rebroadcast the distance
 *calculation request
 */
class DistanceFwTimer : public TimerHandler {
 public:
    DistanceFwTimer (DensityAgent* a) :TimerHandler(), a_(a) {};
 protected:
    virtual void expire(Event *);
    DensityAgent* a_;
 };

/*==========IN DELEGATION==================================*/

/*
 *When the delegation timer fires, the current IN identifies the next IN and
 *then delegates the election operations
 */
class DelegationTimer : public TimerHandler {
 public:
    DelegationTimer (DensityAgent* a) :TimerHandler(), a_(a) {};
 protected:
    virtual void expire(Event *);
    DensityAgent* a_;
 };


/*===============================================================
 *                      BOUNDARIES MAINTENANCE
 *==============================================================*/

/*
 *When this timer fires the node sends its periodical Hello packet
 */
class SendHelloTimer : public TimerHandler {
public:
        SendHelloTimer(DensityAgent* a) : a_(a) {};
        void	expire(Event*);
private:
        DensityAgent    *a_;
        Event	intr;
};

/*
 *When this timer fires, the node checks whether it dense MANET
 *participation conditions still hold (i.e. if the number of neighbors
 *exceeds the requested) or not.
 */
class CheckNeighborTimer : public TimerHandler {
public:
        CheckNeighborTimer(DensityAgent* a) : a_(a) {}
        void	expire(Event*);
private:
        DensityAgent    *a_;
      	Event	intr;
};


/*=====================================================================================================
 *                                    DENSITY AGENT
 *=====================================================================================================*/
class DensityAgent : public Agent {
public:
  DensityAgent(nsaddr_t id);
  int command(int argc, const char*const* argv);
  void recv(Packet*, Handler*);
  Packet* BCastPacket();
  void timeout(int);
  
  /*Boundaries Identification protocol*/
  void start();
  void recvDensityDiscovery(Packet* pkt);
  void discover();
  void advertise();

  /*Manager Election protocol*/
  /*Farthest nodes identification*/
  void send_calc_req();
  void forwardforward_DstReq();
  void distance_reply(int nHops, int srcid);
  void recvDistReq(Packet* pkt);
  void recvDistRep(Packet* pkt);
  /*Next IN delegation*/
  /*The delegation logic*/
  void delegate();
  void delegate_lessThanBest(hdr_den_deleg_calc*);
  void delegate_higherThanBest(hdr_den_deleg_calc*);
  void delegate_equalBest(hdr_den_deleg_calc*);
  void delegate_first();
  /*End*/
  int deleg_next (int unexplored_num_, int unexplored_list_[]);
  void delegate(int next);
  /*Backwards if the current delegate is not able to see the next*/
  void forward_deleg(Packet* pkt);
  void delegate(int next, int bestVal, int bestNode, int bestNum, int worstVal, int*, int, int*, int, bool, int);
  /*Manager Election Helper functions*/
  int createExplored (hdr_den_deleg_calc* hdr, int* explored);
  int createUnExplored(int* unexplored, int* source, int source_num, int*explored, int explored_num);
  int merge(int* temp, int* front, int front_num, int* unexplored, int unexplored_num);
  void set_FwN();
  /*This function is used to check who is the next hop toward the next delegate*/
  int find_responsible(int);
  

  /*Boundaries Maintenance protocol*/
  void sendHello();
  void recvHello(Packet*);
  void nb_purge();
  /*Boundaries Maintenance helper functions*/ 
  NeighborEntry nb_lookup(int);
  void nb_insert(int);
  void nb_update(int, double);
  void nb_delete(int);

  /*Utility function*/
  int belongs(int, int*, int);
 


  //Intervals
  //Delay to advertise my presence to my neighbors
  static int disc_delay_;
  //Intervals and delays to advertise the number of my neighbors
  static int part_adv_interval_;
  static int part_adv_delay_;
  //Limit of neighbors to participate to the hotspot
  static int limit_;
  //The interval to reply
  static int dist_rep_interval_;
  //The interval to forward
  static int dist_req_delay_;
  //The accuracy required in the leader Election
  static float election_accuracy_;
  //The interval requested to receive replies from distant nodes
  static int delegation_interval_;
  //The probability of forwarding a received request for forwarding
  static float forwarding_prob_;
  //The probability that a distant node replies
  static float dist_reply_prob_;
  //The number of max best solutions explored in a row
  static int max_best_;
  //The hello period
  static float hello_period_;
  //The hello delay
  static float hello_delay_;
  //The check period
  static float check_hello_period_;



private:
 
  int part_;
  //Node state (0 in discovery, 1 after broadcast, 3 already sent the DST_REQ, 4 I am the initiator of the Calculus)
  int state_;
  //Node identifier
  int nid_;
  //Number of neighbors detected
  int counter_;

  //List of my neighbors' neighbors number
  NeighborEntry table_[NODENUM];
  //Number of my neighbors' neighbors table received
  int ngbEnNum_;
  //The discovery timer
  DiscoveryTimer disc_timer_;
  //Is this agent the initiator?
  int initiator_;
  //The participation advertisement timer
  PartAdvTimer part_adv_timer_;
  //From who I received the CHAIN_MSG
  int from_;
  //To who I sent last CHAIN_MSG
  int last_to_;
  //IF I AM RUNNING ELECTION TEST OR NOT
  int test_;

  
  
  //The number of neighbors replying the query
  int count_test_;




  //===========================================//
  //==============HELLO PACKETS================//
  //==============TIMERS=======================//
  //The Hello Timer
  SendHelloTimer send_hello_timer_;
  //The periodical neighbor check timer
  CheckNeighborTimer check_neighbor_timer_;

 

  //===========================================//
  //================DISTANT NODE CALC==========//
  //The node which sent us the packet
  int first_forwarder_;
  //The distance timer
  DistanceRepTimer dist_rep_timer_;
  //The distance from the current IN
  int distance_;
  //The distance queror
  int current_IN_;
  //The distance forwarding timer
  DistanceFwTimer dist_fw_timer_;
  //The delegation timer
  DelegationTimer deleg_timer_;

  /*Next attributes contain neighbors that forward messages incoming from all dense MANET BOUNDARY nodes*/
  //The nodes which forwarded to the current node boundary nodes' messages
  int forwarders_[NODENUM/2];
  //The hops of the boundary node that the forwarders forwarded
  int forwarders_hops_[NODENUM/2]; 
  //The number of forwarders
  int forwarders_num_;
  //The node has already been delegate
  int del_;

  /*Next attributes contain node's neighbors that forward messages incoming only from dense MANET FARTHEST nodes*/
  //My Forwarding Neighbors List
  int fwN_[NODENUM/2];
  //The dimension of fwN_
  int fwN_num_;

  /*To avoid using routing protocol, we introduce a simplified routing table:
    this is useful to reach delegates during the election.
    The table contains only delegates "children" in the delegation tree.*/
  RTEntry rt_[NODENUM/2];
  //The dimension of the routing table
  int rt_num_;
  
  
  /*BOUNDARY MAINTENANCE*/

  /*The list of the node's neighbors (built by exploiting Hello packets)*/
  NeighborEntry neighbors_[NODENUM];
  //Current neighbors2 lenght
  int neighbors_num_;

  int neighbors2_[NODENUM];
  
  
  //The value of the present node
  int INvalue_;
  
  //How many ask for distance propagation I received:
  //the more I received, the less probability there is I will forward them
  int dist_rec_num_;


  /*Temporary variable to mainain the packet arrived*/
  Packet* delegation_packet_;


};


/*=====================================================================================================
 *                                  DENSITY PARTNER
 *=====================================================================================================*/

/*
 *This Density Partner helps filtering broadcast messages
 */

class DensityPartner : public Connector {
public:
        DensityPartner();
        void recv(Packet *p, Handler *h);

protected:
	int command(int argc, const char*const*argv);
        ns_addr_t here_;
	int densityagent_;
        int mask_;
        int shift_;
};



#endif
